/* libargus/libargus.h
 *
 * Client-side interface to argusd.
 *
 * Copyright (C) 2004
 * Chris Lance   <cal6838@covad.net>>
 * Zorica Majdov <zmajdov@virtuoso.com>
 * Andy Goth     <unununium@openverse.com>
 *
 * This code is available under the GNU General Public License; see COPYING. */

#ifndef SEEN_LIBARGUS_H
#define SEEN_LIBARGUS_H

struct argus;     typedef struct argus     argus_t;
struct argus_shm; typedef struct argus_shm argus_shm_t;

#if 0

/* Summary: */

typedef int (*argus_cb_t)(argus_t* ctx, char* sensor, int value, void* extra);

typedef struct argus {
    int         socket_fd;          /* fd for argusd socket, or -1.         */
    char*       socket_path;        /* Path to argusd socket node.          */
    argus_shm_t* shm;               /* Pointer to the shm, or NULL.         */
    char**      sensor_names;       /* List of sensors being monitored.     */
    int*        sensor_values;      /* Current values for sensors.          */
    argus_cb_t* sensor_callbacks;   /* Callback functions for sensors.      */
    void**      callback_data;      /* void*'s passed to callback funcs     */
    int         sensor_count;       /* Len of names, CBs and values lists.  */
    argus_cb_t* packet_callback;    /* Callback function for data packets.  */
    void*       packet_cb_data;     /* void* passed to packet_callback()    */
    int         loop_exit;          /* If true, argus_loop() exits.         */
    int         version;            /* LCD between client & server versions */
    
    unsigned short packet_length;
    unsigned short packet_size;
    unsigned char* packet;
} argus_t;

typedef struct argus_shm {
    char** names;                   /* Names of all sensors.                */
    int*   values;                  /* Current values for all sensors.      */
    int    count;                   /* Total number of sensors.             */
    char*  shm_data;                /* Pointer to the shm data area.        */
} argus_shm_t;

int argus_connect    (argus_t* ctx, char* socket_path);
int argus_disconnect (argus_t* ctx);
int argus_listen     (argus_t* ctx, char** sensors);
int argus_callback   (argus_t* ctx, char* sensor, void* data, argus_cb_t cb);
int argus_loop       (argus_t* ctx);
int argus_loop_exit  (argus_t* ctx);
int argus_data_wait  (argus_t* ctx, int timeout);
int argus_data_read  (argus_t* ctx);
int argus_sensors    (argus_t* ctx, char*** sensors);
int argus_open_shm   (argus_t* ctx);
int argus_close_shm  (argus_t* ctx);

#endif

/* Typedef: argus_cb_t
 *   Sensor/packet callback function, executed every time a sensor changes or a
 *   data packet is received, depending on how it is registered.
 *
 * Argument: argus_t* ctx
 *   Argus connection context.  This is most useful for getting at sensor
 *   values.
 *
 * Argument: char* sensor
 *   Name of the sensor triggering the callback.  In the case of packet
 *   callbacks, this argument is NULL.
 *
 * Argument: int value
 *   Current value of the sensor.  For packet callbacks, this is -1.
 *
 * Argument: void* extra
 *   Extra parameter, as passed to argus_callback() when registering this
 *   callback function.  This can be NULL if not needed or can point to a
 *   struct containing additional data necessary for the callback function.
 *
 * Return value: int
 *   On success, 0.  On failure, -1, and errno is set.  If the callback
 *   function returns error, argus_read() returns the same error, which causes
 *   argus_loop() to exit.  To exit the enclosing argus_loop() without
 *   returning an error, use argus_loop_exit(). */
typedef int (*argus_cb_t)(argus_t* ctx, char* sensor, int value, void* extra);

struct argus {
    /* fd used for reading from and writing to the argusd socket. */
    int socket_fd;

    /* Path to the argusd socket node.  Can be either a filesystem path or a
     * 108-byte name in the abstract socket namespace. */
    char* socket_path;

    /* Shared memory segment for faster client <--> server communication. */
    argus_shm_t* shm;

    /* List of names of the sensors currently being listened to.  This array is
     * both NULL-terminated and length-counted by sensor_count. */
    char** sensor_names;

    /* The most recently-read values for all of the sensors.  This is a
     * parallel array of sensor_names, so the value of the sensor named
     * sensor_names[5] is stored at sensor_values[5]. */
    int* sensor_values;

    /* List of callback functions for each sensor.  A NULL function pointer
     * indicates that a given sensor currently has no callback.  Again, this is
     * parallel with sensor_names and sensor_values. */
    argus_cb_t* sensor_callbacks;

    /* List of callback data for each sensor callback function.  This array is
     * parallel with sensor_callbacks[].  */
    void** callback_data;

    /* The length of the sensor_names, sensor_values, and sensor_callback
     * arrays. */
    int sensor_count;

    /* Callback to be executed every time a new data packet is received.  Since
     * data packets are only sent whenever a sensor changes, this function
     * won't be called needlessly.  Unlike registering this callback with
     * every sensor individually, it'll only be called once per packet.
     * Packet callbacks are executed before sensor callbacks. */
    argus_cb_t packet_callback;

    void* packet_cb_data;

    /* If true, argus_loop() exits. */
    int loop_exit;
    
    /* Highest protocol version understood by both client and server. */
    int version;

    unsigned short packet_length;
    unsigned short packet_size;
    unsigned char* packet;
};

struct argus_shm {
    /* Array containing the names of all sensors.  This array is parallel with
     * the above values array, so for example, the sensor named names[2] has
     * value values[2].  Since the sensor names cannot be changed without
     * restarting the argusd server, this array need only be read once, or
     * you can use argus_sensors() to get the same information.  The final
     * element of this array is NULL. */
    char** names;

    /* Array containing the current values of all sensors in the system.
     * There's no locking mechanism, so beware that a single read may result
     * in values taken from multiple samplings.  If you require atomicity,
     * use the socket mechanism instead, or bug me to add semaphores. */
    int*   values;

    /* Total number of sensors. */
    int    count;

    /* Pointer to the start of the shm.  The above names and values fields
     * point into this array. */
    char*  shm_data;
};

/* Function: argus_connect()
 *   Opens a connection to argusd.  This function doesn't return until the
 *   connection is successfully opened or a fatal error is encountered.
 *
 * Argument: argus_t* ctx
 *   Argus connection context.  This argus_t* is passed to all other argus
 *   functions.
 *
 * Argument: char* socket_path
 *   Path to the socket node to connect to.  If a C-style null-terminated
 *   string, this gives the path to a UNIX socket in the filesystem.  If the
 *   first byte is \0, this is a 108-byte string giving the name of a socket in
 *   Linux's abstract socket namespace.
 *
 * Return value: int
 *   On success, 0.  On failure, -1, and errno is set. */
extern int argus_connect(argus_t* ctx, char* socket_path);

/* Function: argus_disconnect()
 *   Closes a connection to argusd.
 *
 * Argument: argus_t* ctx
 *   Argus connection context describing the connection to close.  If this call
 *   succeeds, and if the argus shm is not currently open, then the argus_t is
 *   deallocated and therefore this pointer becomes invalid and should not be
 *   passed to any argus functions.  On the other hand, if the argus shm is
 *   open, then deallocation is deferred to argus_close_shm().
 *
 * Return value: int
 *   On success, 0.  On failure, -1, and errno is set. */
extern int argus_disconnect(argus_t* ctx);

/* Function: argus_listen()
 *   Registers interest in the given list of sensors.  This function returns
 *   after completing all negotiation with argusd.  Values for all requested
 *   sensors are provided by the next call to argus_data_read(), and at that
 *   time, all callback functions will be triggered.
 *
 * Argument: argus_t* ctx
 *   Argus connection context.
 *
 * Argument: char** sensors
 *   List of names of sensors to listen to.  This list should be terminated by
 *   a NULL pointer.  The names in this list are copied to ctx->sensor_names.
 *   It is an error to specify a single sensor more than once.
 *
 * Return value: int
 *   On success, 0.  On failure, -1, and errno is set.  If argusd was able to
 *   find sensors[i] (which is equal to ctx->sensor_names[i]), then
 *   ctx->sensor_values[i] is -1, otherwise -2. */
extern int argus_listen(argus_t* ctx, char** sensors);

/* Function: argus_callback()
 *   Registers a callback function for a given named sensor.  This callback
 *   function is executed every time the named sensor's value changes.
 *
 * Argument: argus_t* ctx
 *   Argus connection context.
 *
 * Argument: char* sensor
 *   Name of the sensor for which values changes trigger the callback
 *   function.  Note that a single sensor can have only one callback function.
 *
 * Argument: void* data
 *   Extra argument to pass to the callback function.
 *
 * Argument: argus_cb_t cb
 *   Pointer to the function to be called when the sensor changes.  Since only
 *   one callback can be registered for each sensor, this replaces the previous
 *   value.  Passing NULL for this argument disables the callback.
 *
 * Return value: int
 *   On success, 0.  On failure, -1, and errno is set. */
extern int argus_callback(argus_t* ctx, char* sensor, void* data,
argus_cb_t cb);

/* Function: argus_loop()
 *   Enters an event loop which waits for incoming data from the argusd server
 *   and continually processes it, executing callback functions as necessary.
 *   If a callback function returns negative, this function returns with error,
 *   preserving the errno value set by the callback function.
 *
 * Argument: argus_t* ctx
 *   Argus connection context.
 *
 * Return value: int
 *   On success, 0.  On failure, -1, and errno is set. */
extern int argus_loop(argus_t* ctx);

/* Function: argus_loop_exit()
 *   If called from a sensor callback, causes the enclosing argus_loop() to
 *   exit, returning control to its caller.
 *
 * Argument: argus_t* ctx
 *   Argus connection context.
 *
 * Return value: int
 *   On success, 0.  On failure, -1, and errno is set. */
extern int argus_loop_exit(argus_t* ctx);

/* Function: argus_data_wait()
 *   Halts program execution until new data is available from the argusd
 *   server.  Doesn't return until a data packet arrives, the timeout expires,
 *   or an error occurs (such as disconnection).
 *
 * Argument: argus_t* ctx
 *   Argus connection context.
 *
 * Argument: int timeout
 *   Maximum amount of time in milliseconds argus_data_wait() will block
 *   program execution waiting for data to arrive.  If negative, no timeout is
 *   set, and argus_data_wait() will not return until data arrives or an error
 *   occurs.  If zero, argus_data_wait() will return virtually immediately,
 *   reporting whether or not new data is available.
 *
 * Return value: int
 *   On success, 0.  On failure, -1, and errno is set.
 *
 *   XXX: There must be a way to tell whether the connection breaks (or other
 *   connection error), whether data has arrived, or whether the timeout has
 *   expired. */
extern int argus_data_wait(argus_t* ctx, int timeout);

/* Function: argus_data_read()
 *   Reads and processes a packet of data from the argusd server, executing
 *   callback functions as necessary.  After this function returns with
 *   success, new data is available.  If it is undesirable for the program to
 *   wait potentially forever for new data to arrive, use argus_data_wait() and
 *   a timeout.
 *
 * Argument: argus_t* ctx
 *   Argus connection context.
 *
 * Return value: int
 *   On success, 0.  On failure, -1, and errno is set. */
extern int argus_data_read(argus_t* ctx);

/* Function: argus_sensors()
 *   Requests a list of all available sensors from the argusd server.  This
 *   list can also be obtained by reading ctx->shm->names.
 *
 * Argument: argus_t* ctx
 *   Argus connection context.
 *
 * Argument: char*** sensors
 *   The address of a char** in which to store the list of names.  Beware: this
 *   is a triple pointer.  The char** allocated by this function is of the same
 *   basic form as the char** passed to argus_listen(), that is, it's an array
 *   of char*'s terminated by a NULL pointer.  It's certainly possible to call
 *   this function then call argus_listen() with the results in order to listen
 *   to all sensors on the server.  The char** is allocated in such a way that
 *   it can be entirely freed with a single call to free() with no need for
 *   iteration.
 *
 * Return value: int
 *   On success, the number of sensors available through the argusd server.  On
 *   failure, -1, and errno is set. */
extern int argus_sensors(argus_t* ctx, char*** sensors);

/* Function: argus_open_shm()
 *   Opens a read-only argus_shm_t connected to the argusd server.  At most one
 *   argus_shm_t per server connection is possible.  After this call succeeds,
 *   ctx->shm points to the argus_shm_t.
 *
 * Argument: argus_t* ctx
 *   Argus connnection context through which the argus_shm_t is to be opened.
 *
 * Return value: int
 *   On success, 0.  On failure, -1, and errno is set. */
extern int argus_open_shm(argus_t* ctx);

/* Function: argus_close_shm()
 *   Closes a previously-opened argus_shm_t connected to the argusd server.
 *
 * Argument: argus_t* ctx
 *   Argus connection context describing the argus_shm_t to close.
 *   argus_disconnect() may have already been called; if so, this call
 *   deallocates the argus_t, invalidating this pointer.
 *
 * Return value: int
 *   On success, 0.  On failure, -1, and errno is set. */
extern int argus_close_shm(argus_t* ctx);

#endif

/* vim: set ts=4 sts=4 sw=4 tw=80 et: */
/* EOF */
