/* libgoth/vector.h
 *
 * A general linear array structure.
 *
 * Copyright (C) 2003, 2004
 * Andy Goth     <unununium@openverse.com>
 *
 * This code is available under the GNU General Public License; see COPYING. */

#ifndef SEEN_LIBGOTH_VECTOR_H
#define SEEN_LIBGOTH_VECTOR_H

/* A function for sorting vectors.  Returns negative if 'a' sorts before 'b',
 * zero if 'a' and 'b' sort equally, and positive if 'a' sorts after 'b'. */
typedef int(*vector_sort_func_t)(const void* a, const void* b);

/* This vector type should work for any kind of data.  It iss optimized for
 * frequent traversal and infrequent insertion.  The vector never shrinks, only
 * grows, so it sacrifices a bit of memory for speed, but then again the queue
 * is the same way. */
typedef struct vector {
    char* data;             /* Pointer to start of array data. */
    int count;              /* Number of elements in array. */
    int capacity;           /* Maximal number of elements in array. */
    int sizeof_elem;        /* Number of bytes per element. */

    /* Sort function used by vector_insert_sorted(). */
    vector_sort_func_t sort_func;
} vector_t;

/* Creates a new empty vector with room for 'size' elements, each 'sizeof_elem'
 * bytes in length.  Its address is stored at *'b'. */
extern int vector_new(vector_t** v, int size, int sizeof_elem);

/* Deallocates vector 'v'. */
extern int vector_del(vector_t* v);

/* Resizes vector 'v' to have room for 'new_size' elements.  Returns 'new_size'
 * on success and an error code on failure. */
extern int vector_resize(vector_t* v, int new_size);

/* Adds 'data', a pointer to an array of 'count' elements, to the vector 'v'
 * at position 'pos'.  Negative positions are relative to the end of the
 * vector.  The vector is resized to hold all incoming data.  Returns the
 * number of elements written on success and an error code on failure. */
extern int vector_insert(vector_t* v, int pos, void* data, int count);

/* Copies 'count' elements from vector 'v' to the 'data' array.  Copied data is
 * removed from 'v'.  Returns the number of elements written on success and an
 * error code on failure.  Data is copied from position 'pos'.  If 'pos' is
 * negative, it is relative to the end of the vector. */
extern int vector_pop(vector_t* v, int pos, void* data, int count);

/* Copies 'count' elements from vector 'v' to the 'data' array.  Returns the
 * number of elements written on success and an error code on failure.  Data
 * is copied from position 'pos'.  If 'pos' is negative, it is relative to
 * the end of the vector. */
extern int vector_peek(vector_t* v, int pos, void* data, int count);

/* Sets *'data' to point to a linear array of 'count' elements taken from the
 * vector 'v' starting at position 'pos'.  Returns the number of elements in
 * the array (may be more than 'count') on success and an error code on
 * failure.  If 'pos' is negative, it is relative to the end of the vector. */
extern int vector_index(vector_t* v, int pos, void* data, int count);

/* Throws away up to 'count' elements from vector 'v' at position 'pos'.
 * Returns the number of discarded elements on success and an error code on
 * failure.  If 'pos' is negative, it is relative to the end of the vector. */
extern int vector_discard(vector_t* v, int pos, int count);

/* Returns the number of elements currently stored in vector 'v'. */
extern int vector_count(vector_t* v);

/* Returns the current size in elements of vector 'v'. */
extern int vector_size(vector_t* v);

/* Sets vector 'v''s sort function to 'sort_func'. */
extern int vector_set_sort_func(vector_t* v, vector_sort_func_t sort_func);

/* Sorts vector 'v'. */
extern int vector_sort(vector_t* v);

/* Searches for an element 'data' in vector 'v'.  Stores the index of the first
 * occurrence in *'first' and one plus the index of the last occurrence in
 * *'last'.  If 'data' is not found, 'first' equals 'last' and both point to
 * the position in which 'data' would be inserted to maintain sort order.  On
 * success, returns the number of occurrences of 'data' (which is equal to
 * *'last' minus *'first'), and on failure returns a negative error code.
 * 'first' and/or 'last' can be NULL, but this causes vector_find_sorted()
 * to always return 0 on success rather than the number of occurrences. */
extern int vector_find_sorted(vector_t* v, void* data, int* first, int* last);

/* Inserts an element 'data' into vector 'v'.  The insertion position is
 * determined by sort_func().  Returns the insertion position on success and
 * a negative error code on failure. */
extern int vector_insert_sorted(vector_t* v, void* data);

#endif

/* ======================================================================== */

/* vim: set ts=4 sts=4 sw=4 tw=80 et: */
/* EOF */
