/* ss-core/com/serial.c
 *
 * Facilitaties communication with devices on RS232 serial ports.
 *
 * Copyright (C) 2003, 2004
 * Zorica Majdov <zmajdov@virtuoso.com>
 * Andy Goth     <unununium@openverse.com>
 *
 * This code is available under the GNU General Public License; see COPYING. */

/* === Header files. ====================================================== */

#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <termios.h>

#include "ss-core/com/serial.h"
#include "ss-core/lib/device.h"
#include "ss-core/lib/error.h"

/* === Type definitions, static variables, and static prototypes. ========= */

/* Parameters useful for RS232 serial devices. */
typedef struct serial {
    int data_bits;                  /* CS5, CS6, CS7, CS8. */
    int stop_bits;                  /* 0 = 1 stop bit; CSTOPB = 2 stop bits. */
    int parity;                     /* 0 = no parity; PARENB, PARODD. */
    int baud_rate;                  /* ..., B300, ..., B57600, ... */
    int sw_flow_ctrl;               /* 0 = off; IXON | IXOFF = on. */
    int hw_flow_ctrl;               /* 0 = off; CRTSCTS = on. */
    struct termios old_termios;     /* Saved termios configuration. */
} serial_t;

/* Closes previously-opened serial port 'd'. */
static int serial_close(device_t* d);

/* === Non-static function definitions. =================================== */

/* Opens a serial port and connects it to device 'd'.  The serial port is
 * configured according to the given line parameters. */
int serial_open(device_t* d, int data_bits, int stop_bits, int parity,
                int baud_rate, int sw_flow_ctrl, int hw_flow_ctrl)
{
    int ret, fd;
    struct termios current;
    serial_t* s;
    if (d == NULL) return -E_INVAL;

    /* Set hardware configuration for the dev_t. */
    s = malloc(sizeof(*s));
    if (s == NULL) return -E_MEMORY;
    d->on_readable = NULL;
    d->on_writable = NULL;
    d->on_cleanup_com = serial_close;
    d->com_state = s;
    
    /* Open the serial port. */
    fd = open(d->path, O_RDWR);
    if (fd == -1) {ret = -E_ERRNO; goto error1;}
    device_connect_fd(d, fd, fd);

    /* Get current settings. */
    if (tcgetattr(fd, &current) != 0) {
        int saved_errno = errno;
        error_print(-E_ERRNO, "serial_open()", "Failure sensing serial "
                    "line settings.");
        errno = saved_errno;
        ret = -E_ERRNO; goto error2;
    }

    /* Fill the serial_t struct. */
    s->old_termios  = current;
    s->data_bits    = data_bits;    s->stop_bits    = stop_bits;
    s->parity       = parity;       s->baud_rate    = baud_rate;
    s->sw_flow_ctrl = sw_flow_ctrl; s->hw_flow_ctrl = hw_flow_ctrl;

    /* Configure the port. */
    current.c_iflag &= ~(IXON   | IXANY | IXOFF | ISTRIP | INLCR |
                         ICRNL  | IGNCR | IUCLC);
    current.c_oflag &= ~(OLCUC  | ONLCR | OCRNL | ONOCR  | ONLRET);
    current.c_lflag &= ~(ICANON | ECHO  | ECHOE | ECHONL | ISIG);
    current.c_cflag &= ~(CBAUD  | PARENB);
#ifdef CBAUDEX
    current.c_cflag &= ~(CBAUDEX);
#endif
    current.c_iflag |= (sw_flow_ctrl);
    current.c_cflag |= (baud_rate    | parity       | data_bits    |
                        stop_bits    | hw_flow_ctrl | CLOCAL       );
    current.c_cc[VTIME] = 0;
    current.c_cc[VMIN]  = 1;

    /* Commit the settings. */
    if (tcsetattr(fd, TCSANOW, &current) != 0) {
        int saved_errno = errno;
        error_print(-E_ERRNO, "serial_open()", "Failure configuring "
                    "serial line.");
        errno = saved_errno;
        ret = -E_ERRNO; goto error2;
    }

    /* Success. */
    return E_OK;

error2: device_connect_fd(d, -1, -1);
        close(fd);
error1: free(s);
        return ret;
}

/* === Static function definitions. ======================================= */

/* Closes previously-opened serial port 'd'. */
static int serial_close(device_t* d)
{
    int ret = E_OK;

    serial_t* s;
    if (d == NULL) return -E_INVAL;
    s = d->com_state;
    if (s == NULL) return -E_INVAL;

    if (tcsetattr(d->r_fd, TCSANOW, &s->old_termios) != 0) {
        int saved_errno = errno;
        error_print(-E_ERRNO, "serial_close()", "Failure restoring "
                    "serial line.");
        errno = saved_errno;
        ret = -E_ERRNO;
    }

    close(d->r_fd);
    device_connect_fd(d, -1, -1);
    free(s);

    return ret;
}

/* ======================================================================== */

/* vim: set ts=4 sts=4 sw=4 tw=80 et: */
/* EOF */
